/*********************************************************************************
*                    "Game of life" mit LED-Matrix-8x8
*                    =================================
*                             Uwe Berger; 2009
*
*
* Spielregeln fuer Game of life (http://de.wikipedia.org/wiki/Conways_Spiel_des_Lebens)
* -------------------------------------------------------------------------------------
* Jede Zelle hat 8 Nachbarn (im Raster); Zellen ausserhalb der Begrenzung gelten
* als tot
* --> tote Zellen
*   1. eine tote Zelle mit genau 3 lebenden Nachbarn lebt in der Folgegeneration
* --> lebende Zellen
*   2. lebende Zellen mit weniger als 2 lebenden Nachbarn sterben
*   3. lebende Zellen mit 2 oder 3 lebenden Nachbarn leben weiter
*   4. lebende Zellen mit mehr als 3 lebenden Nachbarn sterben
*
*
* Programmkurzbeschreibung
* ------------------------
* Die Anfangsbilder werden mit zwei verschiedene Methoden generiert (umschaltbar mit,
* am MC angeschlossene Tasten (siehe Defines)):
*
* --> "Statischer" Mode
* Im  PROGMEM sind einige bekannte Ausgangspopulationen des Game of life abgelegt, die 
* man mit der entsprechenden Taste abrufen kann.
*
* --> Zufallsmode
* Damit es nicht langweilig wird, koennen die Ausgangspopulationen auch dem Zufall
* ueberlassen werden. Mit der vorgesehenen Taste wird zum einen diese Mode eingeschaltet,
* bei jedem neuen Tastendruck die Groesse des Feldes fuer das Ausgangsbild erhoeht und 
* dann dieses Feld zufaellig mit lebenden Zellen gefuellt.
*
* In beiden Modi erfolgt die Berechnung der nchsten Generation automatisch zyklisch mit
* einer kleinen Pause dazwischen. Bleibt eine Population in einer Sachgasse stecken (alle 
* Zellen sind gestorben oder die Konstellation der lebenden Zelle bleibt zwischen den 
* Generationen gleich) erfolgt eine Neuinitialisierung der Ausgangspopulation.
*
*
* ---------
* Have fun!
*
**********************************************************************************/
#ifndef F_CPU
	#define F_CPU 1000000
#endif

#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <stdlib.h>


//*** Hardwaredefinitionen ***
// Matrix-Zeilen
#define ROW_0_DDR		DDRC
#define ROW_0_PORT		PORTC
#define ROW_0			PC0
#define ROW_1_DDR		DDRC
#define ROW_1_PORT		PORTC
#define ROW_1			PC1
#define ROW_2_DDR		DDRC
#define ROW_2_PORT		PORTC
#define ROW_2			PC2
#define ROW_3_DDR		DDRC
#define ROW_3_PORT		PORTC
#define ROW_3			PC3
#define ROW_4_DDR		DDRC
#define ROW_4_PORT		PORTC
#define ROW_4			PC4
#define ROW_5_DDR		DDRC
#define ROW_5_PORT		PORTC
#define ROW_5			PC5
#define ROW_6_DDR		DDRD
#define ROW_6_PORT		PORTD
#define ROW_6			PD0
#define ROW_7_DDR		DDRD
#define ROW_7_PORT		PORTD
#define ROW_7			PD1
// Matrix-Spalten
#define COL_0_DDR		DDRD
#define COL_0_PORT		PORTD
#define COL_0			PD3
#define COL_1_DDR		DDRD
#define COL_1_PORT		PORTD
#define COL_1			PD4
#define COL_2_DDR		DDRD
#define COL_2_PORT		PORTD
#define COL_2			PD5
#define COL_3_DDR		DDRD
#define COL_3_PORT		PORTD
#define COL_3			PD6
#define COL_4_DDR		DDRD
#define COL_4_PORT		PORTD
#define COL_4			PD7
#define COL_5_DDR		DDRB
#define COL_5_PORT		PORTB
#define COL_5			PB0
#define COL_6_DDR		DDRB
#define COL_6_PORT		PORTB
#define COL_6			PB1
#define COL_7_DDR		DDRB
#define COL_7_PORT		PORTB
#define COL_7			PB2

// Taster
#define TASTER_ON_OFF_DDR	DDRD
#define TASTER_ON_OFF_PORT	PORTD
#define TASTER_ON_OFF_PINX  PIND
#define TASTER_ON_OFF_PIN	PD2

#define TASTER_LEFT_DDR		DDRB
#define TASTER_LEFT_PORT	PORTB
#define TASTER_LEFT_PINX	PINB
#define TASTER_LEFT_PIN		PB7

#define TASTER_RIGHT_DDR	DDRB
#define TASTER_RIGHT_PORT	PORTB
#define TASTER_RIGHT_PINX	PINB
#define TASTER_RIGHT_PIN	PB6

// Tastencodes
#define ON_OFF	1
#define LEFT	2
#define RIGHT	3

#define DEBOUNCE_MS			10

#define PRELOAD_T0 217

volatile uint8_t row = 0;
volatile uint8_t anim_timer = 0;
uint8_t idx_cells = 0;
uint8_t random_mode = 0;
uint8_t random_size = 3;

// "Anfangszellkulturen"
//--> 0-Laser
const uint8_t c1[] PROGMEM = {
							0b11000000,
							0b11000000,
							0b00110000,
							0b00110000,
							0b00000000,
							0b00000000,
							0b00000000,
							0b00000000
							};
//--> 2-Laser
const uint8_t c2[] PROGMEM = {
							0b11000000,
							0b10100000,
							0b00000000,
							0b00101000,
							0b00011000,
							0b00000000,
							0b00000000,
							0b00000000
							};
							
//--> Blinker
const uint8_t c3[] PROGMEM = {
							0b01000010,
							0b01000010,
							0b01000010,
							0b00000000,
							0b00000000,
							0b01000010,
							0b01000010,
							0b01000010
							};
//--> Unruhe
const uint8_t c4[] PROGMEM = {
							0b00000000,
							0b00100000,
							0b00101000,
							0b01010000,
							0b00010000,
							0b00000000,
							0b00000000,
							0b00000000
							};
//--> Oktagon
const uint8_t c5[] PROGMEM = {
							0b00011000,
							0b00111100,
							0b01000010,
							0b11000011,
							0b11000011,
							0b01000010,
							0b00111100,
							0b00011000
							};
//--> Gleiter
const uint8_t c6[] PROGMEM = {
							0b01000000,
							0b00100000,
							0b11100000,
							0b00000000,
							0b00000000,
							0b00000000,
							0b00000000,
							0b00000000
							};
const uint8_t *cells[] PROGMEM = {c1, c2, c3, c4, c5, c6};

// Anzeigepuffer
volatile uint8_t led[8] = {
							0b00000000,
							0b00000000,
							0b00000000,
							0b00000000,
							0b00000000,
							0b00000000,
							0b00000000,
							0b00000000
							};
							
// temporaeres Ergebnisfeld der naechsten Generation
uint8_t temp_led[8];

// "Zugvektoren" fuer die Ermittlung der angrenzenden Felder
const uint8_t vx[8] = {-1, -1, -1,  0, 0,  1, 1, 1};
const uint8_t vy[8] = {-1,  0,  1, -1, 1, -1, 0, 1};

//************************************************************************
void matrix_clear(void) {
	// alle Zeilen aus (Low)
	ROW_0_PORT &= ~(1<<ROW_0);
	ROW_1_PORT &= ~(1<<ROW_1);
	ROW_2_PORT &= ~(1<<ROW_2);
	ROW_3_PORT &= ~(1<<ROW_3);
	ROW_4_PORT &= ~(1<<ROW_4);
	ROW_5_PORT &= ~(1<<ROW_5);
	ROW_6_PORT &= ~(1<<ROW_6);
	ROW_7_PORT &= ~(1<<ROW_7);
	// alle Spalten aus (High)
	COL_0_PORT |= (1<<COL_0);
	COL_1_PORT |= (1<<COL_1);
	COL_2_PORT |= (1<<COL_2);
	COL_3_PORT |= (1<<COL_3);
	COL_4_PORT |= (1<<COL_4);
	COL_5_PORT |= (1<<COL_5);
	COL_6_PORT |= (1<<COL_6);
	COL_7_PORT |= (1<<COL_7);
}

//************************************************************************
uint8_t get_led(uint8_t x, uint8_t y)
{
	return (bit_is_set(led[y], 7-x));
}

//************************************************************************
ISR(TIMER0_OVF_vect)
{
	TCNT0 = PRELOAD_T0;	
	// Matrix zuruecksetzen
	matrix_clear();
	// naechste Zeile/Spalte ermitteln
	row++;
	if (row > 7) row = 0;
	// Spalten setzen
	if (get_led(0, row)) COL_0_PORT &= ~(1<<COL_0);
	if (get_led(1, row)) COL_1_PORT &= ~(1<<COL_1);
	if (get_led(2, row)) COL_2_PORT &= ~(1<<COL_2);
	if (get_led(3, row)) COL_3_PORT &= ~(1<<COL_3);
	if (get_led(4, row)) COL_4_PORT &= ~(1<<COL_4);
	if (get_led(5, row)) COL_5_PORT &= ~(1<<COL_5);
	if (get_led(6, row)) COL_6_PORT &= ~(1<<COL_6);
	if (get_led(7, row)) COL_7_PORT &= ~(1<<COL_7);
	// naechste Zeile ein (Low)
	switch (row) {
		case 0: ROW_0_PORT |= (1 << ROW_0); break;
		case 1: ROW_1_PORT |= (1 << ROW_1); break;
		case 2: ROW_2_PORT |= (1 << ROW_2); break;
		case 3: ROW_3_PORT |= (1 << ROW_3); break;
		case 4: ROW_4_PORT |= (1 << ROW_4); break;
		case 5: ROW_5_PORT |= (1 << ROW_5); break;
		case 6: ROW_6_PORT |= (1 << ROW_6); break;
		case 7: ROW_7_PORT |= (1 << ROW_7); break;
	}
	// Zaehler fuer automatische Weiterschaltung der Generationen
	anim_timer++;
}

//************************************************************************
ISR(INT0_vect)
{
	// deklariert, um wieder aus Sleep-Mode zu kommen
}

//************************************************************************
void my_delay_ms(uint16_t ms)
{
	uint16_t i;
	for (i=0; i<ms; i++) _delay_ms(1);
}

//************************************************************************
uint8_t scankey(volatile uint8_t *pinx, uint8_t pin) 
{
	if (bit_is_clear(*pinx, pin)) {
		my_delay_ms(DEBOUNCE_MS);
		if (bit_is_clear(*pinx, pin)) return 1;
	}
	return 0;
}

//************************************************************************
uint8_t getkey(void) 
{
	if (scankey(&TASTER_ON_OFF_PINX, TASTER_ON_OFF_PIN))	return ON_OFF;
	if (scankey(&TASTER_LEFT_PINX, TASTER_LEFT_PIN))		return LEFT;
	if (scankey(&TASTER_RIGHT_PINX, TASTER_RIGHT_PIN))		return RIGHT;
	return 0;
}

//************************************************************************
void clear_all(void)
{
	uint8_t i;
	for (i=0; i<8; i++) led[i] = 0;
}

//************************************************************************
void clear_temp_led(void)
{
	uint8_t i;
	for (i=0; i<8; i++) temp_led[i] = 0;
}

//************************************************************************
void set_led(uint8_t x, uint8_t y, uint8_t val)
{
	x = 7-x;
	if (val) {led[y] |= (1 << x);} else {led[y] &= ~(1 << x);}
}

//************************************************************************
void copy_temp_led_to_led(void)
{
	uint8_t i;
	for (i=0; i<8; i++) led[i] = temp_led[i];
}

//************************************************************************
// (Ausgangsbilder aus PROGMEM in Anzeigepuffer kopieren)
void copy_cells_to_led(uint8_t idx)
{
	const uint8_t *p;
	uint8_t i;
	p = (const uint8_t*)(pgm_read_word(&(cells[idx])));
	for (i=0; i<8; i++) led[i] = pgm_read_byte(p+i);
}

//************************************************************************
void set_temp_led(uint8_t x, uint8_t y, uint8_t val)
{
	x = 7-x;
	if (val) {temp_led[y] |= (1 << x);} else {temp_led[y] &= ~(1 << x);}
}

//************************************************************************
uint8_t count_lifing_neighbors(uint8_t x, uint8_t y)
{
	uint8_t count = 0;
	uint8_t i;
	int8_t nx, ny;
	// Anzahl der lebenden Nachbarn zaehlen
	for (i=0; i<8; i++ ) {
		nx=(int8_t)x+vx[i];
		ny=(int8_t)y+vy[i];
		if ((nx>-1)&&(ny>-1)&&(nx<8)&&(ny<8)) if (get_led(nx, ny)) count++;
	}
	return count;
}

//************************************************************************
// Returnwert=1, wenn sich das berechnete Bild nicht mehr vom Ausgangsbild
// unterscheidet, ansonsten Returnwert=0
uint8_t next_generation(void)
{
	uint8_t x, y, count;
	// Zwischenfeld loeschen
	clear_temp_led();
	// alle Zellen des Spielfeldes durchgehen
	for (y=0; y<8; y++)
		for (x=0; x<8; x++)
			if (get_led(x, y)) {
				// ist eine lebende Zelle
				// Anzahl lebende Nachbarn ermitteln
				count = count_lifing_neighbors(x, y);
				// Regel 2
				if (count < 2) set_temp_led(x, y, 0);
				// Regel 3
				if ((count == 2)||(count == 3)) set_temp_led(x, y, 1);
				// Regel 4
				if (count > 3) set_temp_led(x, y, 0);
			} else {
				// ist eine tote Zelle
				// Regel 1
				if (count_lifing_neighbors(x, y) == 3) set_temp_led(x, y, 1);
			};
	// hat sich das Bild ueberhaupt geaendert? (Variablen x, y werden recycled)
	y=0;
	for (x=0; x<8; x++) if (led[x]!=temp_led[x]) y=1; 
	if (!y) return 1;
	// Zwischenfeld in Bildspeicher kopieren
	copy_temp_led_to_led();
	// ansonsten fertisch!
	return 0;
}

//*********************************************************************
// Returnwert 0 oder 1 als Zufallswert
uint8_t my_rand(void)
{
	return rand() % 2;
}

//*********************************************************************
// seed-Wert fuer srand() generieren; Quelle: 
// http://www.roboternetz.de/wissen/index.php/Zufallszahlen_mit_avr-gcc
unsigned int get_seed(void)
{
	uint16_t seed = 0;
	uint16_t *p = (uint16_t*) (RAMEND+1);
	extern uint16_t __heap_start;
	while (p >= &__heap_start + 1)
		seed ^= * (--p);
	return seed;
}

//************************************************************************
void generate_random_life(void)
{
	uint8_t x, y;
	for (y=0; y<random_size; y++) 
		for (x=0; x<random_size; x++) set_led(x, y, my_rand());
}

//************************************************************************
//************************************************************************
//************************************************************************
int main(void)
{
	uint8_t x, y;
	// Hardware initialisieren...
	// ... Analogcomperator ausschalten (Strom sparen!)
	ACSR = 0x80;		
	//... Ausgaenge
	ROW_0_DDR |= (1<<ROW_0);
	ROW_1_DDR |= (1<<ROW_1);
	ROW_2_DDR |= (1<<ROW_2);
	ROW_3_DDR |= (1<<ROW_3);
	ROW_4_DDR |= (1<<ROW_4);
	ROW_5_DDR |= (1<<ROW_5);
	ROW_6_DDR |= (1<<ROW_6);
	ROW_7_DDR |= (1<<ROW_7);
	COL_0_DDR |= (1<<COL_0);
	COL_1_DDR |= (1<<COL_1);
	COL_2_DDR |= (1<<COL_2);
	COL_3_DDR |= (1<<COL_3);
	COL_4_DDR |= (1<<COL_4);
	COL_5_DDR |= (1<<COL_5);
	COL_6_DDR |= (1<<COL_6);
	COL_7_DDR |= (1<<COL_7);
	// ... Eingaenge
	TASTER_ON_OFF_PORT &= ~(1 << TASTER_ON_OFF_PIN);
	TASTER_LEFT_PORT   &= ~(1 << TASTER_LEFT_PIN);
	TASTER_RIGHT_PORT  &= ~(1 << TASTER_RIGHT_PIN);
	// ... Timer initialisieren
    TCNT0   = PRELOAD_T0;							// Preloadwert Timer0 setzen
	TCCR0 |= (1 << CS01)|(1 << CS00);				// 64
	TIMSK  |= (1 << TOIE0);							// Timer0-OVF einschalten
	// ... externe Interrupts
	//GICR  |= (1<<INT0);          					// externer Interrupt an INT0
	// ... Interrupts einschalten
	sei();
	// Zufallsgenerator initialisieren
	srand(get_seed());					
	// erstes Bild
	copy_cells_to_led(idx_cells);
	// Hauptschleife	
	while (1){
		// Tasten abfragen/reagieren
		switch (getkey()) {
			case ON_OFF:	{	// **** On/Off-Schalter ****
								// Timer0-OVF ausschalten
								TIMSK  &= ~(1 << TOIE0);
								// Ports entsprechend "loeschen"...
								matrix_clear();
								// Pause damit nicht gleich INT0 
								my_delay_ms(500);
								// INT0 einschalten
								GICR  |= (1<<INT0);
								// Sleep-Mode einstellen
								set_sleep_mode(SLEEP_MODE_PWR_DOWN);	
								// Sleep-Mode einschalten
								sleep_mode();
								// INT0 ausschalten		
								GICR  &= ~(1<<INT0);
								//	Timer0-OVF einschalten				
								TIMSK  |= (1 << TOIE0);
								// Pause wg. sofort wieder getkey()
								my_delay_ms(500);
								break;
							}
			case LEFT:		{
								// --> statische Ausgangsbilder (siehe oben)
								// Zufallsmode ausschalten
								random_mode = 0;
								// naechstes Anfangsbild
								idx_cells++;
								if (idx_cells >= (sizeof(cells)/2)) idx_cells=0;
								copy_cells_to_led(idx_cells);
								my_delay_ms(500);								
								break;
							}
			case RIGHT:		{
								// --> zufaellige Ausgangsbilder 
								// Zufallsmode einschalten
								random_mode = 1;
								// Groesse des Anfangsbildes erhoehen
								random_size++;
								if (random_size>8) random_size = 3;
								// Anfangsbildfeld kurz anzeigen
								clear_all();
								for (y=0; y<random_size; y++) 
									for (x=0; x<random_size; x++) set_led(x, y, 1);
								my_delay_ms(500);
								clear_all();
								// zufaellige Pkte. im Anzeigefeld erzeugen
								generate_random_life();
								break;
							}
		}
		// automatisch naechste Generation berechnen?
		if (anim_timer > 100) {
			if (next_generation()){ 
				// wenn berechnetes Bild in einer "Sackgasse" dann neues Ausgangsbild...
				if (random_mode) {
					clear_all();
					//... mittels Zufall ermitteln
					generate_random_life();
				} else {
					//... gleiches Ausgangsbild nochmal laden
					copy_cells_to_led(idx_cells);
				}
			}
			anim_timer = 0;
		}
	}
}
